@@ -1,13 +1,18 @@
 import math
 from abc import ABC
-from typing import Callable, Union, Tuple, List
+from enum import Enum
+from typing import Callable, Union, Tuple, List, Optional
 
 import torch
 from transformers import CLIPTokenizer, CLIPTextModel
 
-__all__ = ["EmbeddingsProvider"]
+__all__ = ["EmbeddingsProvider", "DownweightMode"]
 
 
+class DownweightMode(Enum):
+    REMOVE = 0  # Remove downweighted tokens from the token sequence (shifts all subsequent tokens)
+    MASK = 1,   # Default: Leave tokens in-place but mask them out using attention masking
+
 class BaseTextualInversionManager(ABC):
     def expand_textual_inversion_token_ids_if_necessary(self, token_ids: List[int]) -> List[int]:
         raise NotImplementedError()
@@ -16,12 +21,14 @@ class BaseTextualInversionManager(ABC):
 class EmbeddingsProvider:
 
     def __init__(self,
-                tokenizer: CLIPTokenizer,
-                text_encoder: CLIPTextModel,
-                textual_inversion_manager: BaseTextualInversionManager = None,
-                dtype_for_device_getter: Callable[[torch.device], torch.dtype] = lambda device: torch.float32,
-                truncate: bool = True
-                ):
+                 tokenizer: CLIPTokenizer,
+                 text_encoder: CLIPTextModel,
+                 textual_inversion_manager: BaseTextualInversionManager = None,
+                 dtype_for_device_getter: Callable[[torch.device], torch.dtype] = lambda device: torch.float32,
+                 truncate: bool = True,
+                 padding_attention_mask_value: int = 1,
+                 downweight_mode: DownweightMode = DownweightMode.MASK
+                 ):
         """
         `tokenizer`: converts strings to lists of int token ids
         `text_encoder`: convert lists of token ids to embedding tensors
@@ -29,11 +36,14 @@ class EmbeddingsProvider:
         `dtype_for_device_getter`: callback that returns an appropriate dtype for the requested device. if unset, defaults to torch.float32.
         `truncate`: if True, truncate inputs to the maximum length specified by the tokenizer. if False, returns
                     tensors that may be longer than the maximum length (but will always be an integer multiple of maximum length)
+        `padding_attention_mask_value`: Value to write into the attention mask for padding tokens. Stable Diffusion needs 1.
         """
         self.tokenizer = tokenizer
         self.text_encoder = text_encoder
         self.textual_inversion_manager = textual_inversion_manager
         self.truncate_to_model_max_length = truncate
+        self.padding_attention_mask_value = padding_attention_mask_value
+        self.downweight_mode = downweight_mode
 
         # by default always use float32
         self.get_dtype_for_device = dtype_for_device_getter
@@ -66,6 +76,8 @@ class EmbeddingsProvider:
 
         :param text_batch: A list of fragments of text to which different weights are to be applied.
         :param fragment_weights_batch: A list of weights, one for each entry in `fragments`.
+        :param should_return_tokens: If True, return a tuple of (embeddings, tokens), otherwise just return embeddings.
+        :param device: Where to put the constructed tensor(s)
         :return: A tensor of shape `[1, 77, token_dim]` containing weighted embeddings where token_dim is 768 for SD1
                     and 1280 for SD2
         """
@@ -89,8 +101,8 @@ class EmbeddingsProvider:
             # closer the resulting embedding is to an embedding for a prompt that simply lacks this fragment.
 
             # handle weights >=1
-            tokens, per_token_weights = self.get_token_ids_and_expand_weights(fragments, weights, device=device)
-            base_embedding = self.build_weighted_embedding_tensor(tokens, per_token_weights)
+            tokens, per_token_weights, mask = self.get_token_ids_and_expand_weights(fragments, weights, device=device)
+            base_embedding = self.build_weighted_embedding_tensor(tokens, per_token_weights, mask)
 
             # this is our starting point
             embeddings = base_embedding.unsqueeze(0)
@@ -101,15 +113,31 @@ class EmbeddingsProvider:
             # with the embeddings tensors that have the words, such that if the weight of a word is 0.5, the resulting
             # embedding will be exactly half-way between the unweighted prompt and the prompt with the <1 weighted words
             # removed.
-            # eg for "mountain:1 man:0.5", intuitively the "man" should be "half-gone". therefore, append an embedding
+            # e.g. for "mountain:1 man:0.5", intuitively the "man" should be "half-gone". therefore, append an embedding
             # for "mountain" (i.e. without "man") to the already-produced embedding for "mountain man", and weight it
             # such that the resulting lerped embedding is exactly half-way between "mountain man" and "mountain".
+            fragment_token_index_ranges = self._get_token_ranges_for_fragments(tokens.tolist(), fragments)
+
             for index, fragment_weight in enumerate(weights):
                 if fragment_weight < 1:
-                    fragments_without_this = fragments[:index] + fragments[index+1:]
-                    weights_without_this = weights[:index] + weights[index+1:]
-                    tokens, per_token_weights = self.get_token_ids_and_expand_weights(fragments_without_this, weights_without_this, device=device)
-                    embedding_without_this = self.build_weighted_embedding_tensor(tokens, per_token_weights)
+                    if self.downweight_mode == DownweightMode.MASK:
+                        fragment_start_token_id, fragment_end_token_id = fragment_token_index_ranges[index]
+                        # mask out this fragment
+                        mask_without_fragment = mask.clone()
+                        mask_without_fragment[fragment_start_token_id:fragment_end_token_id+1] = 0
+                        if not self.truncate_to_model_max_length:
+                            # but don't mask chunk-delimiting eos/bos markers
+                            mask_without_fragment[0::self.tokenizer.model_max_length] = 1
+                            mask_without_fragment[self.tokenizer.model_max_length-1::self.tokenizer.model_max_length] = 1
+                        embedding_without_this = self.build_weighted_embedding_tensor(tokens, per_token_weights, mask_without_fragment)
+                    else:
+                        fragments_without_this = fragments[0:index] + fragments[index+1:]
+                        weights_without_this = weights[0:index] + weights[index+1:]
+                        tokens_without_fragment, per_token_weights_without_fragment, mask_without_fragment = \
+                            self.get_token_ids_and_expand_weights(fragments_without_this, weights_without_this, device=device)
+                        embedding_without_this = self.build_weighted_embedding_tensor(tokens_without_fragment,
+                                                                                      per_token_weights_without_fragment,
+                                                                                      mask_without_fragment)
 
                     embeddings = torch.cat((embeddings, embedding_without_this.unsqueeze(0)), dim=1)
                     # weight of the embedding *without* this fragment gets *stronger* as its weight approaches 0
@@ -123,7 +151,7 @@ class EmbeddingsProvider:
                     #        1.0 at PI/4, and
                     #        inf at PI/2
                     # -> tan((1-weight)*PI/2) should give us ideal lerp weights
-                    epsilon = 1e-9
+                    epsilon = 1e-5
                     fragment_weight = max(epsilon, fragment_weight) # inf is bad
                     embedding_lerp_weight = math.tan((1.0 - fragment_weight) * math.pi / 2)
                     # todo handle negative weight?
@@ -183,7 +211,8 @@ class EmbeddingsProvider:
 
         return result
 
-    def get_token_ids_and_expand_weights(self, fragments: List[str], weights: List[float], device: str) -> (torch.Tensor, torch.Tensor):
+    def get_token_ids_and_expand_weights(self, fragments: List[str], weights: List[float], device: str
+                                         ) -> (torch.Tensor, torch.Tensor, torch.Tensor):
         '''
         Given a list of text fragments and corresponding weights: tokenize each fragment, append the token sequences
         together and return a padded token sequence starting with the bos marker, ending with the eos marker, and padded
@@ -193,7 +222,8 @@ class EmbeddingsProvider:
         :param fragments: Text fragments to tokenize and concatenate. May be empty.
         :param weights: Per-fragment weights (i.e. quasi-CFG scaling). Values from 0 to inf are permitted. In practise with SD1.5
                         values >1.6 tend to produce garbage output. Must have same length as `fragment`.
-        :return: A tuple of tensors `(token_ids, weights)`. `token_ids` is ints, `weights` is floats, both have shape `[self.max_length]`.
+        :return: A tuple of tensors `(token_ids, weights, mask)`. `token_ids` is ints, `weights` is floats, `mask` is
+                        ints, all have shape `[self.max_length]`.
         '''
         if len(fragments) != len(weights):
             raise ValueError(f"lengths of text and fragment_weights lists are not the same ({len(fragments)} != {len(weights)})")
@@ -205,7 +235,7 @@ class EmbeddingsProvider:
         per_fragment_token_ids = self.get_token_ids(fragments, include_start_and_end_markers=False)
         all_token_ids: List[int] = []
         all_token_weights: List[float] = []
-        #print("all fragments:", fragments, weights)
+        # print("all fragments:", fragments, weights)
         for this_fragment_token_ids, weight in zip(per_fragment_token_ids, weights):
             # append
             all_token_ids += this_fragment_token_ids
@@ -215,7 +245,7 @@ class EmbeddingsProvider:
         return self._chunk_and_pad_token_ids(all_token_ids, all_token_weights, device=device)
 
     def _chunk_and_pad_token_ids(self, token_ids: List[int], token_weights: List[float], device: str
-                                 ) -> Tuple[torch.Tensor, torch.Tensor]:
+                                 ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
 
         remaining_token_ids = token_ids
         remaining_token_weights = token_weights
@@ -223,6 +253,7 @@ class EmbeddingsProvider:
 
         all_token_ids = []
         all_token_weights = []
+        all_masks = []
         while True:
             # each chunk must leave room for bos/eos
             chunk_token_ids = remaining_token_ids[0:chunk_length_without_eos_bos_markers]
@@ -235,12 +266,16 @@ class EmbeddingsProvider:
             # (typically self.max_length == 77)
             chunk_token_ids = [self.tokenizer.bos_token_id] + chunk_token_ids + [self.tokenizer.eos_token_id]
             chunk_token_weights = [1.0] + chunk_token_weights + [1.0]
+            chunk_mask = [1.0] * len(chunk_token_ids)
+
             pad_length = self.max_token_count - len(chunk_token_ids)
             chunk_token_ids += [self.tokenizer.pad_token_id] * pad_length
             chunk_token_weights += [1.0] * pad_length
+            chunk_mask += [self.padding_attention_mask_value] * pad_length
 
             all_token_ids += chunk_token_ids
             all_token_weights += chunk_token_weights
+            all_masks += chunk_mask
 
             if self.truncate_to_model_max_length or len(remaining_token_ids) == 0:
                 break
@@ -249,17 +284,24 @@ class EmbeddingsProvider:
         all_per_token_weights_tensor = torch.tensor(all_token_weights,
                                                     dtype=self.get_dtype_for_device(self.text_encoder.device),
                                                     device=device)
-        #print(f"assembled all_token_ids_tensor with shape {all_token_ids_tensor.shape}")
-        return all_token_ids_tensor, all_per_token_weights_tensor
+        all_masks = torch.tensor(all_masks, dtype=torch.long, device=device)
+        # print(f"assembled all_token_ids_tensor with shape {all_token_ids_tensor.shape}")
+        return all_token_ids_tensor, all_per_token_weights_tensor, all_masks
 
 
-    def build_weighted_embedding_tensor(self, token_ids: torch.Tensor, per_token_weights: torch.Tensor) -> torch.Tensor:
+    def build_weighted_embedding_tensor(self,
+                                        token_ids: torch.Tensor,
+                                        per_token_weights: torch.Tensor,
+                                        attention_mask: Optional[torch.Tensor] = None) -> torch.Tensor:
         """
         Build a tensor that embeds the passed-in token IDs and applies the given per_token weights
 
         :param token_ids: A tensor of shape `n*[self.max_length]` containing token IDs (ints) where n is some arbitrary
-            integer (i.e. n==1 for shorter prompts, or it may be >1 if there are more than max_length tokens in the original prompt)
+            integer (i.e. n==1 for shorter prompts, or it may be >1 if there are more than max_length tokens in the
+            original prompt)
         :param per_token_weights: A tensor containing weights (floats), with the same shape as token_ids
+        :param attention_mask: A tensor containing a mask (ints), with the same shape as token_ids, where 1 means use
+            the corresponding token and 0 means ignore the corresponding token.
 
         :return: A tensor of shape `[1, token_ids.shape[0], token_dim]` representing the requested weighted embeddings
             where `token_dim` is 768 for SD1 and 1280 for SD2.
@@ -279,10 +321,15 @@ class EmbeddingsProvider:
         chunk_size = self.max_token_count
         while chunk_start_index < token_ids.shape[0]:
             next_chunk_start_index = chunk_start_index+chunk_size
-            chunk_token_ids = token_ids[chunk_start_index:next_chunk_start_index]
             chunk_per_token_weights = per_token_weights[chunk_start_index:next_chunk_start_index]
+            chunk_token_ids = token_ids[chunk_start_index:next_chunk_start_index].unsqueeze(0)
+            chunk_attention_mask = (
+                attention_mask[chunk_start_index:next_chunk_start_index].unsqueeze(0)
+                if attention_mask is not None
+                else None
+            )
 
-            z = self.text_encoder(chunk_token_ids.unsqueeze(0), return_dict=False)[0]
+            z = self.text_encoder(chunk_token_ids, attention_mask=chunk_attention_mask, return_dict=False)[0]
             batch_weights_expanded = chunk_per_token_weights.reshape(
                 chunk_per_token_weights.shape + (1,)).expand(z.shape).to(z)
 
@@ -296,3 +343,67 @@ class EmbeddingsProvider:
             chunk_start_index += chunk_size
 
         return weighted_z
+
+    def _get_token_ranges_for_fragments(self, chunked_and_padded_token_ids: List[int], fragments: List[str]) -> List[Tuple[int, int]]:
+        """
+        Match token id sequences for the strings in `fragments` with token id sequences in `chunked_and_padded_token_ids`,
+         taking into account any eos and bos markers marking `self.tokenizer.max_model_length` sized chunks.
+
+        Returns a list of tuples indicating start and end indices of each fragment's corresponding token id sequence in
+         `chunked_and_padded_token_ids`.
+        """
+        per_fragment_token_ids = self.get_token_ids(fragments, include_start_and_end_markers=False)
+        fragment_start = 0
+
+        corresponding_indices = []
+        for fragment_index, fragment_token_ids in enumerate(per_fragment_token_ids):
+            if len(fragment_token_ids) == 0:
+                corresponding_indices.append((None, None))
+                fragment_start += 1
+                continue
+            if self.truncate_to_model_max_length and fragment_start >= self.tokenizer.model_max_length - 1:
+                break
+            # find the start
+            while True:
+                if fragment_start >= len(chunked_and_padded_token_ids)-1:
+                    if self.truncate_to_model_max_length:
+                        fragment_start = len(chunked_and_padded_token_ids)-1
+                        break
+                    else:
+                        raise RuntimeError(
+                            f"couldn't find start of token sequence for fragment at index {fragment_index} '{fragments[fragment_index]}'")
+                if chunked_and_padded_token_ids[fragment_start] == fragment_token_ids[0]:
+                    break
+                fragment_start += 1
+            # step through
+            fragment_end = fragment_start
+            fragment_relative_index = 0
+            while True:
+                if fragment_end >= len(chunked_and_padded_token_ids)-1:
+                    if self.truncate_to_model_max_length:
+                        fragment_end = len(chunked_and_padded_token_ids)-1
+                        break
+                    else:
+                        raise RuntimeError(
+                            f"couldn't find end of token sequence for fragment at index {fragment_index} '{fragments[fragment_index]}'")
+                if not self.truncate_to_model_max_length and (
+                        chunked_and_padded_token_ids[fragment_end] == self.tokenizer.eos_token_id
+                        or chunked_and_padded_token_ids[fragment_end] == self.tokenizer.bos_token_id
+                ):
+                    # bos/eos: chunk boundaries
+                    fragment_end += 1
+                elif chunked_and_padded_token_ids[fragment_end] == fragment_token_ids[fragment_relative_index]:
+                    # matching token
+                    fragment_relative_index += 1
+                    if fragment_relative_index == len(fragment_token_ids):
+                        break
+                    fragment_end += 1
+                else:
+                    raise RuntimeError(
+                        f"token sequence mismatch for fragment at index {fragment_index} '{fragments[fragment_index]}':"
+                        f"expected {fragment_token_ids}, found {chunked_and_padded_token_ids[fragment_start:fragment_end + 1]}")
+
+            corresponding_indices.append((fragment_start, fragment_end))
+            fragment_start = fragment_end + 1
+
+        return corresponding_indices
